#! /usr/bin/env python

# RAMulator
# Copyright (C) 2006 Sergej Chodarev

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from optparse import OptionParser

norm_instrs = ['read', 'write', 'load', 'store', 'add', 'sub', 'mul', 'div']
jumps = ['jmp', 'jz', 'jnz']
instructions = norm_instrs + jumps + ['halt']
IMMEDIATE, REGISTER, POINTER = range(3)

class RAMSyntaxError(Exception):
	pass
class NotInitializedRegisterError(Exception):
	pass

def parse(source):
	code = []
	labels = {}
	label = None
	for line in source:
		line_label, instr, op = parse_line(line)
		if line_label != None:
			label = line_label
		if instr != None:
			code.append((instr, op, line))
			if label != None:
				labels[label] = len(code)-1
				label = None
	
	return code, labels

def parse_line(line):
	label = None
	instr = None
	op = None
	line = line.strip().lower()
	
	# Separate label
	split_label = line.split(':', 1)
	if len(split_label) == 2:
		label = split_label[0].strip()
		line = split_label[1]
	
	words = line.split(None, 1)
	if len(words) > 0:
		if words[0] not in instructions:
			raise RAMSyntaxError
		instr = words[0]
		if instr in norm_instrs:
			if len(words) < 2:
				raise RAMSyntaxError
			if words[1][0] == '=':
				if instr in ('read', 'store'):
					raise RAMSyntaxError
				op = (IMMEDIATE, int(words[1][1:], 10))
			elif words[1][0] == '*':
				op = (POINTER, int(words[1][1:], 10))
			else:
				op = (REGISTER, int(words[1], 10))
		elif instr in jumps:
			if len(words) < 2:
				raise RAMSyntaxError
			op = words[1]
	return label, instr, op

class RAM(object):
	def __init__(self, source):
		self.reg = {}
		self.pc = 0
		self.code, self.labels = parse(source)
				
	def run(self, stepping):
		while self.pc >= 0:
			instr, op, line = self.code[self.pc]
			if stepping:
				command = raw_input(
					'  ' + line[:-1] + '; (Enter to step, r to run)> ')
				if command in ('r', 'R'):
					stepping = False
			getattr(self, instr)(op)
			if stepping:
				print '  #', self.reg
	
	def get_op_value(self, op):
		if op[0] == IMMEDIATE:
			return op[1]
		elif op[0] == REGISTER:
			try:
				return self.reg[op[1]]
			except KeyError:
				raise NotInitializedRegisterError
		elif op[0] == POINTER:
			try:
				return self.reg[self.reg[op[1]]]
			except KeyError:
				raise NotInitializedRegisterError
	
	def get_op_register(self, op):
		if op[0] == REGISTER:
			return op[1]
		elif op[0] == POINTER:
			try:
				return self.reg[op[1]]
			except KeyError:
				raise NotInitializedRegisterError
	
	def read(self, op):
		r = self.get_op_register(op)
		self.reg[r] = int(raw_input('In: '))
		self.pc += 1
	
	def write(self, op):
		v = self.get_op_value(op)
		print 'Out:', v
		self.pc += 1
	
	def load(self, op):
		v = self.get_op_value(op)
		self.reg[0] = v
		self.pc += 1
	
	def store(self, op):
		r = self.get_op_register(op)
		self.reg[r] = self.reg[0]
		self.pc += 1
	
	def add(self, op):
		v = self.get_op_value(op)
		self.reg[0] += v
		self.pc += 1
	
	def sub(self, op):
		v = self.get_op_value(op)
		self.reg[0] -= v
		self.pc += 1
	
	def mul(self, op):
		v = self.get_op_value(op)
		self.reg[0] *= v
		self.pc += 1
	
	def div(self, op):
		v = self.get_op_value(op)
		self.reg[0] /= v
		self.pc += 1
	
	def jmp(self, op):
		self.pc = self.labels[op]
	
	def jz(self, op):
		if self.reg[0] == 0:
			self.pc = self.labels[op]
		else:
			self.pc += 1
	
	def jnz(self, op):
		if self.reg[0] != 0:
			self.pc = self.labels[op]
		else:
			self.pc += 1
	
	def halt(self, op):
		self.pc = -1

if __name__ == '__main__':
	opt_parser = OptionParser(usage='usage: %prog [program]')
	opt_parser.add_option('-s', '--stepping',
		action='store_true', dest='stepping', default=False,
		help='Enable stepping mode')
	options, args = opt_parser.parse_args()
	if len(args) != 1:
		print 'You must specify file with RAM program!'
	else:
		f = file(args[0])
		ram = RAM(f)
		try:
			ram.run(options.stepping)
		except RAMSyntaxError:
			print "RAM program syntax error"
		except NotInitializedRegisterError:
			print "RAM program tries to use not initialized register"
